global function BTBInit

array <entity> playersWantingToBalance

struct PlayerKDData{
    entity player
    float scorekd
}

struct {
	table<entity, float> lastmoved = {}
} file

int matchElapsed = 0
int differenceMax = 1
int waitTime = 4
int waitedTime = 0
float accumulatedSuggestionImbalance = 0.0
float accumulatedStompImbalance = 0.0
float suggestionLimit = 1.5
float stompLimit = 2.0
float voteFraction = 0.45
float relativeScoreDifference = 0.0
int rebalancedHasOccurred = 0
float suggestionTimer = 10.0
int afkThresold = 5
int afkTime = 70
int scoreKDShuffle = 1
float shuffleRandomness = 0.25

enum eAntiAfkPlayerState {
    ACTIVE
    SUSPICIOUS
    AFK
}


void function BTBInit(){
    differenceMax = GetConVarInt( "btb_difference_max" )
    waitTime = GetConVarInt( "btb_wait_time" )
    suggestionLimit = GetConVarFloat( "btb_suggestion_limit" )
    stompLimit = GetConVarFloat( "btb_stomp_limit" )
    voteFraction = GetConVarFloat( "btb_vote_fraction" )
    afkThresold = GetConVarInt( "btb_afk_threshold" )
    afkTime = GetConVarInt( "btb_afk_time" )
    scoreKDShuffle = GetConVarInt( "btb_skd_shuffle" )

    FSU_RegisterCommand( "balance","\x1b[113m" + FSU_GetString("FSU_PREFIX") + "balance\x1b[0m If enough players vote, teams and scores will be rebalanced", "btb", BTB_BalanceVote, ["bal"] )
    AddCallback_OnPlayerKilled( OnKill )
    AddCallback_OnClientConnected( AddPlayerCallbacks )
    AddCallback_OnClientDisconnected( DeletePlayerRecords )
    AddCallback_OnPlayerRespawned( Moved )
    foreach (entity player in GetPlayerArray()){
        AddPlayerCallbacks( player )
    }

    if (differenceMax != 0){
        SetConVarInt( "FSU_BALANCE_TEAMS_MID_GAME", 0 )
    }
    if (scoreKDShuffle == 1){
        SetConVarInt( "FSU_SHUFFLE_TEAMS_ON_MATCH_START", 0 )
        AddCallback_GameStateEnter( eGameState.Prematch, Prematch)
        AddCallback_GameStateEnter( eGameState.Postmatch, Postmatch)
    }

    thread BTBThread()
    print("[BTB] AutoTeamBalance is running")
}


void function Prematch(){
    if (GetPlayerArray().len() > 0 && GetConVarString( "rank_list" ) != ""){
        print("[BTB] Pulling player ranks and shuffling teams")
        array <entity> playersRanked

        foreach( string UID in split( GetConVarString( "rank_list" ), "," ) ){
            foreach( entity player in GetPlayerArray() ){
                if ( UID == player.GetUID() ){
                    playersRanked.append( player )
                }
            }
        }
        foreach( entity player in GetPlayerArray() ){
            if ( playersRanked.find( player ) == -1 ){
                playersRanked.append( player )
            }
        }

        for(int i = 0; i < GetPlayerArray().len(); i++){
            if(!IsEven(i))
                SetTeam(playersRanked[i], TEAM_IMC)
            else
                SetTeam(playersRanked[i], TEAM_MILITIA)
        }

//         while ( GetPlayerArrayOfTeam(TEAM_IMC).len() > 1 + GetPlayerArrayOfTeam(TEAM_MILITIA).len() ){
//             SetTeam( GetPlayerArrayOfTeam(TEAM_IMC)[RandomIntRange( 0, GetPlayerArrayOfTeam(TEAM_IMC).len() - 1 )], TEAM_MILITIA )
//         }
//         while ( GetPlayerArrayOfTeam(TEAM_MILITIA).len() > 1 + GetPlayerArrayOfTeam(TEAM_IMC).len() ){
//             SetTeam( GetPlayerArrayOfTeam(TEAM_MILITIA)[RandomIntRange( 0, GetPlayerArrayOfTeam(TEAM_MILITIA).len() - 1 )], TEAM_IMC )
//         }
    }
}

void function Postmatch(){
    print("[BTB] Saving player ranks to convar")
    array<PlayerKDData> playerRanks = GetPlayersSortedBySkill(GetPlayerArray(), shuffleRandomness)
    string ranklist = ""

    for(int i = 0; i < GetPlayerArray().len(); i++){
        if (ranklist == ""){
            ranklist = playerRanks[i].player.GetUID()
            print("[BTB] Ranked player saved " + playerRanks[i].player.GetPlayerName() )
        }
        else{
            ranklist = ranklist + "," + playerRanks[i].player.GetUID()
            print("[BTB] Ranked player saved " + playerRanks[i].player.GetPlayerName() )
        }
    }

    SetConVarString( "rank_list", ranklist )
}


// Check if playercount balancing is needed on death
void function OnKill( entity victim, entity attacker, var damageInfo ){
    if( waitedTime >= waitTime ){
        if ( differenceMax == 0 || IsFFAGame() || GetPlayerArray().len() == 1 || abs(GetPlayerArrayOfTeam(TEAM_IMC).len() - GetPlayerArrayOfTeam(TEAM_MILITIA).len()) < differenceMax || GameTime_TimeLeftSeconds() < 60 ){
            return
        }

        // Compare victims teams size
        if ( GetPlayerArrayOfTeam(victim.GetTeam()).len() < GetPlayerArrayOfTeam(GetOtherTeam(victim.GetTeam())).len() ){
            return
        }

        // Passed checks, balance the teams
        print("[BTB] The team of " + victim.GetPlayerName() + " has been switched")
        SetTeam( victim, GetOtherTeam( victim.GetTeam() ) )
        Chat_ServerPrivateMessage( victim, "Your team has been switched to balance the game!", false )
        waitedTime = 1
    }
}

// Count votes for !balance, and activate, when enough have voted for it
void function BTB_BalanceVote ( entity player, array < string > args ){
    if ( GetMapName() == "mp_lobby" ){
        Chat_ServerPrivateMessage( player, "Can't balance in lobby", false )
        return
    }

    if ( GetGameState() != eGameState.Playing ){
        Chat_ServerPrivateMessage( player, "Can't balance in this game state", false )
        return
    }

    if ( rebalancedHasOccurred == 1 ){
        Chat_ServerPrivateMessage( player, "Teams have already been balanced!", false )
        return
    }

    if ( GetPlayerArray().len() < 6 ){
        Chat_ServerPrivateMessage( player, "Not enough players for a good rebalance!", false )
        return
    }

    if ( playersWantingToBalance.find( player ) != -1 ){
        Chat_ServerPrivateMessage( player, "You have already voted!", false )
        return
    }

    playersWantingToBalance.append( player )


    int required_players = int ( GetPlayerArray().len() * voteFraction )
    // rounding down when int division, if there's only one player it would dislpay [1/0] normally
    if ( required_players == 0 ){
        required_players = 1
    }

    if ( playersWantingToBalance.len() >= required_players ){
        print("[BTB] Team skill balancing triggered by vote")
        ExecuteStatsBalance(GetPlayerArray())
        Chat_ServerBroadcast( "\x1b[113m[" + playersWantingToBalance.len() + "/" + required_players + "]\x1b[0m players have voted to balance the teams. It has been done!" )
        rebalancedHasOccurred = 1
    }
    else{
        Chat_ServerBroadcast( "\x1b[113m[" + playersWantingToBalance.len() + "/" + required_players + "]\x1b[0m players want to rebalance the teams and scores, use \x1b[113m" + FSU_GetString("FSU_PREFIX") + "balance\x1b[0m to add your vote!" )
    }
}

// Execute balancing by stats
void function ExecuteStatsBalance(array<entity> _players){
    // sort players based on kd
    array<PlayerKDData> playerRanks = GetPlayersSortedBySkill(_players, 0)

    for(int i = 0; i < GetPlayerArray().len(); i++){
        if(!IsEven(i))
            SetTeam(playerRanks[i].player, TEAM_IMC)
        else
            SetTeam(playerRanks[i].player, TEAM_MILITIA)
    }

    //Balance scores
    if (GAMETYPE == "tdm" || GAMETYPE == "ps" || GAMETYPE == "aitdm" || GAMETYPE == "ttdm"){
        AddTeamScore( TEAM_IMC, -GameRules_GetTeamScore(TEAM_IMC))
        AddTeamScore( TEAM_MILITIA, -GameRules_GetTeamScore(TEAM_MILITIA))

        foreach (entity player in GetPlayerArray()){
            if (GAMETYPE == "tdm" || GAMETYPE == "ps" || GAMETYPE == "ttdm"){
                AddTeamScore( player.GetTeam(), player.GetPlayerGameStat(PGS_KILLS))
            }
            else if (GAMETYPE == "aitdm"){
                AddTeamScore( player.GetTeam(), player.GetPlayerGameStat(PGS_ASSAULT_SCORE))

                //Scale team score to account for missing player if teams are uneven
                int imcPlayers = GetPlayerArrayOfTeam(TEAM_IMC).len()
                int militiaPlayers = GetPlayerArrayOfTeam(TEAM_MILITIA).len()
                if (imcPlayers > 0 && militiaPlayers > 0){
                    if (imcPlayers > militiaPlayers){
                        AddTeamScore( TEAM_MILITIA, GameRules_GetTeamScore(TEAM_MILITIA) * (((militiaPlayers + 1) / imcPlayers) - 1 ) )
                    }
                    else if (militiaPlayers > imcPlayers){
                        AddTeamScore( TEAM_IMC, GameRules_GetTeamScore(TEAM_IMC) * (((imcPlayers + 1) / imcPlayers) - 1 ) )
                    }
                }
            }
        }


    }
    else {
        int totalScore = int ( 0.5 * (GameRules_GetTeamScore(TEAM_IMC) + GameRules_GetTeamScore(TEAM_MILITIA)))
        AddTeamScore( TEAM_IMC, totalScore - GameRules_GetTeamScore(TEAM_IMC))
        AddTeamScore( TEAM_MILITIA, totalScore - GameRules_GetTeamScore(TEAM_MILITIA))
    }
}

// Sort players by stats
array<PlayerKDData> function GetPlayersSortedBySkill(array<entity> arr, float randomness){
    array<PlayerKDData> pkdArr
    foreach (entity player in arr) {
        PlayerKDData temp
        temp.player = player
        int deaths =  player.GetPlayerGameStat(PGS_DEATHS)
        int kills =  player.GetPlayerGameStat(PGS_KILLS)
        float kd = 0.0
        int score = 1
        if (GAMETYPE == "tdm" || GAMETYPE == "ps" || GAMETYPE == "ttdm"){
            score = player.GetPlayerGameStat(PGS_KILLS) + 1
        }
        else if (GAMETYPE == "aitdm") {
            score = player.GetPlayerGameStat(PGS_ASSAULT_SCORE) + 1
        }

        if(deaths == 0){
            if (kills == 0){
                kd = 0.5
            }
            else{
                kd = 1.0 * kills / 0.9
            }
        }
        else{
            if (kills == 0){
                kd = 0.8 / deaths
            }
            else{
                kd = 1.0 * kills / deaths
            }
        }

        temp.scorekd = score * kd * RandomFloatRange( 1.0-randomness, 1.0 )
        pkdArr.append(temp)
    }
    pkdArr.sort(PlayerKDDataSort)
    return pkdArr
}

int function PlayerKDDataSort(PlayerKDData data1, PlayerKDData data2)
{
  if ( data1.scorekd == data2.scorekd )
    return 0
  return data1.scorekd < data2.scorekd ? -1 : 1
}



void function BTBThread(){
    while (true){
        wait 10

        // Check for player count imbalance
        if (differenceMax != 0 && !IsFFAGame() && GetPlayerArray().len() > 1 && GameTime_TimeLeftSeconds() > 60){
            print("[BTB] Checking team imbalance")
            int difference = abs(GetPlayerArrayOfTeam(TEAM_IMC).len() - GetPlayerArrayOfTeam(TEAM_MILITIA).len())

            if (difference > differenceMax)
                waitedTime += difference - differenceMax
            else{
                waitedTime = 0
            }
            if (matchElapsed < 4){
                waitedTime = 30
            }
        }

        // Check for score imbalance
        int imcScore = GameRules_GetTeamScore(TEAM_IMC)
        int militiaScore = GameRules_GetTeamScore(TEAM_MILITIA)

        if (suggestionLimit != 0 || stompLimit != 0 ){
            if (matchElapsed > 12 && imcScore > 5 && militiaScore > 5 && rebalancedHasOccurred == 0 && GameTime_TimeLeftSeconds() > 300 && GetPlayerArray().len() > 5){


                print("[BTB] Checking score imbalance")

                if (militiaScore > imcScore){
                    relativeScoreDifference = 1.0 * militiaScore / imcScore
                }
                else{
                    relativeScoreDifference = 1.0 * imcScore / militiaScore
                }

                if (suggestionLimit != 0 && relativeScoreDifference > suggestionLimit){
                    accumulatedSuggestionImbalance += 1.0 + ((relativeScoreDifference - suggestionLimit) * 2.0 )
                    print("[BTB] accumulated suggestion imbalance/threshold: " + accumulatedSuggestionImbalance + "/" + suggestionTimer)
                }
                else{
                    accumulatedSuggestionImbalance = 0.0
                    suggestionTimer = 10.0
                }
                if (accumulatedSuggestionImbalance > suggestionTimer){
                    print("[BTB] Match is uneven, suggesting rebalance")
                    Chat_ServerBroadcast("Looks like this match is uneven, if you'd like to rebalance the teams and their scores, use \x1b[113m" + FSU_GetString("FSU_PREFIX") + "balance\x1b[0m")
                    suggestionTimer += suggestionTimer * 1.2
                }


                if (stompLimit != 0 && relativeScoreDifference > stompLimit){
                    accumulatedStompImbalance += 1.0 + ((relativeScoreDifference - stompLimit) * 1.5)
                    print("[BTB] accumulated stomp imbalance/threshold: " + accumulatedStompImbalance + "/18")
                }
                else{
                    accumulatedStompImbalance = 0.0
                }
                if (accumulatedStompImbalance > 18 && matchElapsed > 30){
                    print("[BTB] Match is very uneven, forcing rebalance")
                    Chat_ServerBroadcast("Very uneven match detected, teams and scores have been automatically rebalanced.")
                    ExecuteStatsBalance(GetPlayerArray())
                    rebalancedHasOccurred = 1
                }
            }
        }

        // Check for AFK players
        if (afkThresold != 0 && GetPlayerArray().len() > afkThresold){
            print("[BTB] Checking for AFK players")
            foreach (entity player in GetPlayerArray()){

                // remove data if player has left match
                if ( !player.IsPlayer() ){
                    DeletePlayerRecords( player )
                    break
                }

                // ignore if player is FSU admin
                if (CheckAdmin(player)){
                    print("[BTB] AFK player is admin, ignoring")
                    break
                }

                switch (GetAfkState(player)){
                    case eAntiAfkPlayerState.SUSPICIOUS:
                        Chat_ServerPrivateMessage( player, "You will soon be kicked for being AFK! MOVE!!!", false )
                        SendHudMessage( player, "YOU ARE AFK! MOVE!!!", -1, 0.35, 255, 80, 80, 255, 0.15, 5, 0.15 )
                        print("[BTB] AFK player has been warned")
                        break

                    case eAntiAfkPlayerState.AFK:
                        print("[BTB] AFK player kicked: " + player.GetPlayerName() + " - " + player.GetUID())
                        ServerCommand("kickid "+ player.GetUID())
                        break
                }
            }
        }
        matchElapsed += 1
    }
}

bool function CheckAdmin( entity player ){
    foreach (string uid in FSU_GetArray("FSU_ADMIN_UIDS")){
        if( player.GetUID() == uid ){
            return true
        }
    }
    return false
}

int function GetAfkState( entity player ){
    float localgrace = afkTime * 1.0
    float warn = afkTime * 0.43
    if ( player in file.lastmoved){
        float lastmove = file.lastmoved[ player ]
        if (Time() > lastmove + (localgrace - warn)){

            if (Time() > lastmove + localgrace){
                return eAntiAfkPlayerState.AFK
            }
            return eAntiAfkPlayerState.SUSPICIOUS
        }
    }
    return eAntiAfkPlayerState.ACTIVE
}

void function Moved( entity player ){
    file.lastmoved[ player ] <- Time()
}

bool function bMoved( entity player ){
    Moved( player )
    return true
}

void function AddPlayerCallbacks( entity player ){
    AddPlayerPressedForwardCallback( player, bMoved )
    AddPlayerPressedBackCallback( player, bMoved )
    AddPlayerPressedLeftCallback( player, bMoved )
    AddPlayerPressedRightCallback( player, bMoved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.JUMP, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.DODGE, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.LEAVE_GROUND, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.TOUCH_GROUND, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.MANTLE, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.BEGIN_WALLRUN, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.END_WALLRUN, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.BEGIN_WALLHANG, Moved )
    AddPlayerMovementEventCallback( player, ePlayerMovementEvents.END_WALLHANG, Moved )
}

void function DeletePlayerRecords( entity player ){
    if (player in file.lastmoved){
        delete file.lastmoved[ player ]
    }
}
